#include <abort.h>
#include <sys/region.h>
#include "engine.h"

static DisplayList* my_clr_cfb;
static DisplayList* my_clr_zb;

void engineConstruct(Engine* engine) {

	vec_Construct(&engine->lights, sizeof(Lux*), ALIGN_DEFAULT);
	vec_Construct(&engine->rspMatrixes, sizeof(Mtx), ALIGN_8B);
	vec_Resize(&engine->rspMatrixes, MAX_NODES);
	dlistConstruct(&engine->dl);

	displaylistConstruct(&engine->rsp_cfb);
	displaylistConstruct(&engine->rsp_zb);
	my_clr_cfb = &engine->rsp_cfb;
	my_clr_zb = &engine->rsp_zb;

	dlistConstruct(&engine->my_dlist);
	dlistInit(&engine->my_dlist);
}

inline void engineSetViewport(Engine* engine, Viewport* viewport) {

	engine->viewport = viewport;
}

inline void engineSetCamera(Engine* engine, CameraTNode* camera) {

	engine->camera = camera;
}

inline void engineSetRenderTarget(Engine* engine, const Surface* renderTarget) {

	engine->renderTarget = renderTarget;
}

inline void engineSetZBuffer(Engine* engine, const Surface* zbuffer) {

	engine->zBuffer = zbuffer;
}

inline Viewport* engineGetViewport(const Engine* engine) {

	return engine->viewport;
}

inline CameraTNode* engineGetCamera(const Engine* engine) {

	return engine->camera;
}

inline const Surface* engineGetRenderTarget(const Engine* engine) {

	return engine->renderTarget;
}

inline const Surface* engineGetZBuffer(const Engine* engine) {

	return engine->zBuffer;
}

inline void engineAddLight(Engine* engine, const Lux* light) {

	vec_PushBack(&engine->lights, &light);
}

void engineRemoveLight(Engine* engine, const Lux* light) {

	if(engine) {}
	if(light) {}
}

static void engineMeshHandler(Engine* engine, MeshTNode* mesh, Mtx* rspMtx) {

	Gfx* meshdl;
	Material* mat;
	const Pair* pair;
	for(uint32_t i=0; i<vec_GetSize(&mesh->mesh.vpairs); i++) {
		pair = vec_GetConstElement(&mesh->mesh.vpairs, i);
		meshdl = pair->gfx;
		mat = pair->mat;
		if(NULL == mat || NULL == meshdl)
			abort();
#if 0
		uint8_t r = mat->color.x*255;
		uint8_t g = mat->color.y*255;
		uint8_t b = mat->color.z*255;
		uint32_t col = (r<<24) + (g<<16) + (b<<8);
#endif
		dlistLoadMatrix(&engine->dl, mesh->tnode.resultMtxF, rspMtx, MATRIX_MODELVIEW);
		
		if(mat->tex) {
			dlistSetRenderMode(&engine->dl, RM_DECAL);
			dlistLoadTexture(&engine->dl, mat->tex);
		} else {
			dlistSetRenderMode(&engine->dl, RM_SHADE);
		}

		dlistJumpToTranslated(&engine->dl, meshdl);
	}
}

void engineSceneTaskF(Engine* engine, TNode* root) {

	Viewport* vp = engineGetViewport(engine);
	CameraTNode* camtn = engineGetCamera(engine);
	Camera* cam = cameratnodeGetCamera(camtn);
	float aspect = viewportGetWidth(vp) / viewportGetHeight(vp);
	float perspNorm;
	matrixPerspective(&engine->projMtxF, &perspNorm, cameraGetFOVY(cam), aspect, cameraGetNear(cam), cameraGetFar(cam));

	dlistNormalizePerspective(&engine->dl, perspNorm);
	dlistLoadMatrix(&engine->dl, &engine->projMtxF, &engine->projMtx, MATRIX_PROJECTION);
	dlistSetTextureScale(&engine->dl, 1.0f, 1.0f);

	Matrix* mtx = tnodeGetMatrix((TNode*)camtn);
	Matrix inv;
	matrixInvert(mtx, &inv);
	tnodeTransform(root, &inv);

	static NodeIterator ni;
	static bool init = true;
	if(init) {
		nodeIteratorConstruct(&ni, (Node*)root);
		init = false;
	}

	uint_t i = 0;
	TNode* tnode;
	while((tnode = (TNode*)nodeIteratorInc(&ni))) {
		Mtx* rspMtx = vec_GetElement(&engine->rspMatrixes, i++);
		tnode->Process(tnode);
		if(NODE_BILLBOARD == tnode->node.type || NODE_MESH == tnode->node.type)
			engineMeshHandler(engine, (MeshTNode*)tnode, rspMtx);
	}
	nodeIteratorReset(&ni);
}

void engineOverlayTaskF(Engine* engine, const float* alpha) {
	if(engine) {}
	if(alpha) {}

	static Vertex rect_vts[] = {
		{{-1.0f, -1.0f, 0.0f}, {0.0f, 1.0f}},
		{{1.0f, -1.0f, 0.0f}, {1.0f, 1.0f}},
		{{1.0f, 1.0f, 0.0f}, {1.0f, 0.0f}},
		{{-1.0f, 1.0f, 0.0f}, {0.0f, 0.0f}},
	};
	static uint32_t rect_ids[] = {
		0, 1, 3,
		1, 2, 3,
	};
	static Mtx mtx_native1;
	static Mtx mtx_native2;
	static Matrix identity;
	static Matrix mtx;
	static Texture tex = TEXTURE_INIT(320, 88, rect_ids, NULL, TEXTURE_FMT_RGBA16);
	matrixIdentity(&identity);
	matrixScale(&mtx, 0.5f, 0.5f, 1.0f);

	dlistSetVertexSource(&engine->dl, &rect_vts, 0);
	dlistSetIndexSource(&engine->dl, &rect_ids);
	dlistLoadMatrix(&engine->dl, &identity, &mtx_native1, MATRIX_PROJECTION);
	dlistLoadMatrix(&engine->dl, &mtx, &mtx_native2, MATRIX_MODELVIEW);

	dlistLoadTexture(&engine->dl, &tex);
	dlistLoadVertexBuffer(&engine->dl, 0, 0, 4);
	dlistLoadIndexBuffer(&engine->dl, 0, 0, 6);

	dlistSetTextureScale(&engine->dl, 1.0f, 1.0f);
	dlistSetRenderMode(&engine->dl, RM_DECAL);
	dlistDrawPrimitives(&engine->dl, PrimTriangle, 2);
#if 0
	static Vtx screenrect_vts[] = {
		{
			{ {-1,-1,0}, 0, {0,0}, {0xFF, 0xFF, 0xFF, 0x7F} }
		},
		{
			{ {1,-1,0}, 0, {0,0}, {0xFF, 0xFF, 0xFF, 0x7F} },
		},
		{
			{ {-1,1,0}, 0, {0,0}, {0xFF, 0xFF, 0xFF, 0x7F} },
		},
		{
			{ {1,1,0}, 0, {0,0}, {0xFF, 0xFF, 0xFF, 0x7F} },
		},
	};

	static Gfx overlay_dl[] = {
		gsDPSetRenderMode(G_RM_XLU_SURF, G_RM_XLU_SURF2),
		gsDPSetCombineMode(G_CC_SHADE, G_CC_SHADE),
		gsSPClearGeometryMode(G_ZBUFFER | G_LIGHTING),
		gsSPSetGeometryMode(G_SHADE | G_SHADING_SMOOTH ),
		gsSPVertex(&screenrect_vts[0], 4, 0),
		gsSP2Triangles(0, 1, 2, 0, 1, 3, 2, 0),
		gsSPEndDisplayList()
	};

	for(uint16_t i=0; i<4; i++) {
		screenrect_vts[i].v.cn[3] = *alpha*0xFF;
	}
	guMtxIdent(&ident);
	DisplayList* dl = sysGetDisplayList();
	dlAddCommand(dl, gsSPForceMatrix(&ident));
	dlAddCommand(dl, gsSPDisplayList(overlay_dl));
#endif
}

void engineImageTaskF(Engine* engine, TaskImageData* data) {
	if(engine) {}
	if(data) {}
#if 0
	static const Vtx screenrect_vts[] = {
		{
			{ {-1,-1,0}, 0, {0,32<<6}, {0xFF, 0xFF, 0xFF, 0xFF} }
		},
		{
			{ {1,-1,0}, 0, {32<<6,32<<6}, {0xFF, 0xFF, 0xFF, 0xFF} },
		},
		{
			{ {-1,1,0}, 0, {0,0}, {0xFF, 0xFF, 0xFF, 0xFF} },
		},
		{
			{ {1,1,0}, 0, {32<<6,0}, {0xFF, 0xFF, 0xFF, 0xFF} },
		},
	};
	static const Gfx image_dl[] = {
		
		gsDPSetRenderMode(G_RM_XLU_SURF, G_RM_XLU_SURF2),
		gsDPSetCombineMode(G_CC_DECALRGBA, G_CC_DECALRGBA),
		gsSPClearGeometryMode(G_ZBUFFER | G_LIGHTING),
		gsSPSetGeometryMode(G_SHADE | G_SHADING_SMOOTH ),
		
		gsSPEndDisplayList()
	};
	guMtxIdent(&ident);
	DisplayList* dl = sysGetDisplayList();
	dlAddCommand(dl, gsSPForceMatrix(&ident));
	dlAddCommand(dl, gsSPTexture(0x8000, 0x8000, 0, G_TX_RENDERTILE, G_ON));
	dlAddCommand(dl, gsSPDisplayList(image_dl));

	TNode* tnode = &data->obj2d->tnode;
	tnodeProcess(tnode);
	Matrix* mtx = tnode->resultMtxF;
	const Texture* texture = object2DGetConstTexture(data->obj2d);
	uint16_t tileheight = TMEM_SIZE/(textureGetBytesPerPixel(texture)*textureGetWidth(texture));
	const uint16_t tilenpix = tileheight*textureGetWidth(texture);
	const void* pixels = textureGetConstPixels(texture);

	u16 x,y;
	Vector2f ul = engineNDC2DC(4*MATRIX_GET(*mtx, 3, 0), 4*MATRIX_GET(*mtx, 3, 1));
	float sth = MATRIX_GET(*mtx, 1, 1)*tileheight;
	int32_t npix = textureGetPixelCount(texture);
	while(npix > 0) {
		if(npix < tilenpix) {
			tileheight = npix/textureGetWidth(texture);
			sth = MATRIX_GET(*mtx, 1, 1)*tileheight;
		}
		dlAddCommand(dl, gsDPLoadTextureBlock(pixels, G_IM_FMT_RGBA, G_IM_SIZ_16b, textureGetWidth(texture), tileheight, 0, G_TX_WRAP, G_TX_WRAP,
				G_TX_NOMASK, G_TX_NOMASK, G_TX_NOLOD, G_TX_NOLOD));
		
		dlAddCommand(dl, gsSPVertex(&screenrect_vts[0], 4, 0));
		x = ul.x;
		y = ul.y + 4*sth;
		dlAddCommand(dl, gsSPModifyVertex(0, G_MWO_POINT_XYSCREEN, x<<16 | y));
		x = ul.x + 4*MATRIX_GET(*mtx, 0, 0)*textureGetWidth(texture);
		y = ul.y + 4*sth;
		dlAddCommand(dl, gsSPModifyVertex(1, G_MWO_POINT_XYSCREEN, x<<16 | y));
		x = ul.x;
		y = ul.y;
		dlAddCommand(dl, gsSPModifyVertex(2, G_MWO_POINT_XYSCREEN, x<<16 | y));
		x = ul.x + 4*MATRIX_GET(*mtx, 0, 0)*textureGetWidth(texture);
		y = ul.y;
		dlAddCommand(dl, gsSPModifyVertex(3, G_MWO_POINT_XYSCREEN, x<<16 | y));
		
		dlAddCommand(dl, gsSPModifyVertex(0, G_MWO_POINT_ST, tileheight<<5));
		dlAddCommand(dl, gsSPModifyVertex(1, G_MWO_POINT_ST, textureGetWidth(texture)<<5<<16 | tileheight<<5));
		dlAddCommand(dl, gsSPModifyVertex(3, G_MWO_POINT_ST, textureGetWidth(texture)<<5<<16));
		dlAddCommand(dl, gsSP2Triangles(0, 1, 2, 0, 1, 3, 2, 0));

		ul.y += 4*sth;
		*(uint16_t**)&pixels += tilenpix;
		npix -= tilenpix;
	}
#endif
}

void engineClearColorbufF(Engine* engine, void* arg) {
	if(arg) {}
	dlistJumpToTranslated(&engine->dl, dlGetHead(my_clr_cfb));
}

void engineClearZbufF(Engine* engine, void* arg) {
	if(arg) {}
	dlistJumpToTranslated(&engine->dl, dlGetHead(my_clr_zb));
}

void engineProcessTask(Engine* engine, const Task* task) {

	const Surface* cbuf = engineGetRenderTarget(engine);
	const Surface* zbuf = engineGetZBuffer(engine);

	static bool init = false;

	if(!init) {
		dlistSetBuffer(&engine->my_dlist, ZBuffer, 0, surfaceGetWidth(zbuf), surfaceGetHeight(zbuf), surfaceGetData(engineGetZBuffer(engine)));
		dlistSetClearZValue(&engine->my_dlist, 1.0f);
		dlistClearBuffers(&engine->my_dlist, ZBuffer);
		dlistEnd(&engine->my_dlist);
		dltransProcess(&engine->dltrans, &engine->my_dlist, my_clr_zb, NULL, NULL, NULL);

		init = true;
	}
	dlistReset(&engine->my_dlist);
	dlistInit(&engine->my_dlist);
	dlistSetBuffer(&engine->my_dlist, ColorBuffer, 0, surfaceGetWidth(cbuf), surfaceGetHeight(cbuf), surfaceGetData(cbuf));
	dlistSetClearColor(&engine->my_dlist, 0.5f, 0.3f, 0.9f, 1.0f);
	dlistClearBuffers(&engine->my_dlist, ColorBuffer);
	dlistEnd(&engine->my_dlist);
	dlResetCounter(my_clr_cfb);
	dltransProcess(&engine->dltrans, &engine->my_dlist, my_clr_cfb, NULL, NULL, NULL);

	dlistReset(&engine->dl);
	dlistInit(&engine->dl);
	dlistSetBuffer(&engine->dl, ColorBuffer, 0, surfaceGetWidth(cbuf), surfaceGetHeight(cbuf), surfaceGetData(cbuf));
	dlistSetBuffer(&engine->dl, ZBuffer, 0, surfaceGetWidth(zbuf), surfaceGetHeight(zbuf), surfaceGetData(engineGetZBuffer(engine)));

	task->f(engine, task->arg);

	dlistSync(&engine->dl);
	dlistEnd(&engine->dl);
}

#if 0
inline Vector2f engineNDC2DC(float x, float y) {

	Vector2f dc = VECTOR2F_SET(x*320, y*200);
	return dc;
}
#endif

