#ifndef DLIST_H
#define DLIST_H

#include <math/matrix.h>
#include <math/vector.h>
#include <texture.h>
#include <types.h>

typedef uintmax_t	name_t;

typedef struct {
	void*	data;
	void*	current;
	size_t	size;
} DList;

typedef struct {
	Vector3f	position;
	Vector2f	texcoords;
} Vertex;

typedef enum {
	PrimTriangle
} DListPrimitiveType;		// is there a way to force enum to be of specific size, say sizeof(uint8_t)?

typedef enum {
	ColorBuffer = 0x01,
	ZBuffer = 0x02,
} DListBufferType;

typedef enum {
	MATRIX_PROJECTION,
	MATRIX_MODELVIEW,
} MatrixType;

typedef enum {
	RM_SHADE,
	RM_DECAL,
} RenderMode;

typedef enum {
	Init,
	SetVertexSource,
	SetIndexSource,
	LoadVertexBuffer,
	LoadIndexBuffer,
	DrawPrimitives,
	SetBuffer,
	ClearBuffers,
	SetClearColor,
	SetClearZValue,
	LoadTexture,
	Jump,
	End,
	SetTextureScale,
	LoadMatrix,
	LoadLight,
	Sync,
	JumpToTranslated,
	SetRenderMode,
	NormalizePerspective,
} CommandName;

typedef struct {
	name_t		name;
} Command_Init;

typedef struct {
	name_t		name;
	void*		data;
	uint8_t		format;
} Command_SetVertexSource;

typedef struct {
	name_t		name;
	void*		data;
} Command_SetIndexSource;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nvts;
} Command_LoadVertexBuffer;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nids;
} Command_LoadIndexBuffer;

typedef struct {
	name_t		name;
	uint8_t		type;
	uint32_t	num;
} Command_DrawPrimitives;

typedef struct {
	name_t		name;
	void*		buf;
	uint16_t	width;
	uint16_t	height;
	uint8_t		buftype;
	uint8_t		format;
} Command_SetBuffer;

typedef struct {
	name_t		name;
	uint8_t		types;
} Command_ClearBuffers;

typedef struct {
	name_t		name;
	float		r, g, b, a;
} Command_SetClearColor;

typedef struct {
	name_t		name;
	float		val;
} Command_SetClearZValue;

typedef struct {
	name_t		name;
	Texture		texture;
} Command_LoadTexture;

typedef struct {
	name_t		name;
	DList*		dl;
} Command_Jump;

typedef struct {
	name_t		name;
} Command_End;

typedef struct {
	name_t		name;
	float		s;
	float		t;
} Command_SetTextureScale;

typedef struct {
	name_t		name;
	Matrix*		mtx;
	Mtx*		mtx_native;
	MatrixType	type;
} Command_LoadMatrix;

typedef struct {
	name_t		name;
	void*		light;
	uint16_t	idx;
} Command_LoadLight;

typedef struct {
	name_t		name;
} Command_Sync;

typedef struct {
	name_t		name;
	void*		dl;
} Command_JumpToTranslated;

typedef struct {
	name_t		name;
	RenderMode	mode;
} Command_SetRenderMode;

typedef struct {
	name_t		name;
	float		scale;
} Command_NormalizePerspective;

void dlistConstruct(DList* dlist);
void dlistDestroy(DList* dlist);
void dlistReset(DList* dlist);
void* dlistGetData(const DList* dlist);
size_t dlistGetSize(const DList* dlist);

void dlistInit(DList* dlist);
void dlistSetVertexSource(DList* dlist, void* vts, uint8_t format);
void dlistSetIndexSource(DList* dlist, void* ids);
void dlistLoadVertexBuffer(DList* dlist, uint32_t src_ofs, uint32_t dst_ofs, uint32_t nvts);
void dlistLoadIndexBuffer(DList* dlist, uint32_t src_ofs, uint32_t dst_ofs, uint32_t nids);
void dlistDrawPrimitives(DList* dlist, uint8_t type, uint32_t num);
void dlistSetBuffer(DList* dlist, uint8_t buftype, uint8_t format, uint16_t width, uint16_t height, void* buf);
void dlistClearBuffers(DList* dlist, uint8_t types);
void dlistSetClearColor(DList* dlist, float r, float g, float b, float a);
void dlistSetClearZValue(DList* dlist, float val);
void dlistLoadTexture(DList* dlist, const Texture* texture);
void dlistJump(DList* dlist, DList* dl);
void dlistEnd(DList* dlist);
void dlistSetTextureScale(DList* dlist, float s, float t);
void dlistLoadMatrix(DList* dlist, Matrix* mtx, Mtx* mtx_native, MatrixType type);
void dlistLoadLight(DList* dlist, void* light, uint16_t idx);
void dlistSync(DList* dlist);
void dlistJumpToTranslated(DList* dlist, void* dl);
void dlistSetRenderMode(DList* dlist, RenderMode mode);
void dlistNormalizePerspective(DList* dlist, float scale);

#endif	/* DLIST_H */

